/* 
 * Copyright (C) 2000-1 Aaron Lav, asl2@pobox.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <vector>
#include <map>
#include <algorithm>

#include <rfftw.h>

#include <gtk--/style.h>
#include <gdk/gdk.h>



#include "soundcard.h"
#include "stuff.h"
#include "vis.h"
#include "guidistvis.h"



GUIDistVis::tag_drawdata::tag_drawdata ()
{
  rgbbuf = 0;

}
void GUIDistVis::tag_drawdata::set_hw (unsigned int width, unsigned int height)
{
  if (rgbbuf)
    free (rgbbuf);
  h = height;
  w = width;
  rgbbuf = static_cast<guchar *>( calloc (1, 3 * h * w));  
}

GUIDistVis::GUIDistVis (PlayParm *pp)
  :  Vis (pp),
     mainwnd(GTK_WINDOW_TOPLEVEL)
{

  drawing_area.add_events(GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);


  drawing_area.expose_event.connect(SigC::slot(this, &GUIDistVis::expose));
  drawing_area.button_release_event.connect(SigC::slot(this,&GUIDistVis::button_release));

  scroller.set_policy (GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  scroller.add_with_viewport (drawing_area);
  mainwnd.add (scroller);
  drawing_area.show ();
  scroller.show ();
  chunk_cnt = 0;

}

void GUIDistVis::set_parms (Parm *p)
{
  printf ("distvis chunk_cnt %d new %d\n", chunk_cnt, p->nchunk);
  if (chunk_cnt != p->nchunk) {
    chunk_cnt = p->nchunk;
    draw_dat.set_hw (chunk_cnt, chunk_cnt);
    mainwnd.hide ();
    // there seems to be no more direct way to change the size of an
    // already displayed window than to hide it, set its default size,
    // and show it.  I am arrogant enough to think that because the
    // user clicked on recompute, the user wants the window resized
    // to show its contents, rather than having to manually resize the
    // window.  Unfortunately, w/ default twm, this means the user must
    //  re-place the window.
    mainwnd.set_default_size (draw_dat.w + 16, draw_dat.h + 16); // XXX pin to screen?
  // or should that be a WM function?
    //    drawing_area.set_default_size (draw_dat.w, draw_dat.h);
  }

  guchar *ptr = draw_dat.rgbbuf;
  for (unsigned int i = 0; i < chunk_cnt; i++) {
    DiffStore ds (ptr + i * chunk_cnt * 3, draw_dat.w);
    p->samples->diffs_to_pixels (&ds,i);
  }
  if (drawing_area.is_realized ()) {
    drawing_area.get_window ().clear_area_e (0,0,0,0);
  }
  mainwnd.show ();
}


GUIDistVis::~GUIDistVis ()
{

}

int GUIDistVis::expose (GdkEventExpose *e)
{
  draw_data& d = draw_dat;
  if (e->area.x > d.w || e->area.y > d.h) // XXX only b/c extra space...
    return TRUE;
  int stride = d.w * 3;
  guchar *pix_ptr = d.rgbbuf + e->area.y * stride + e->area.x * 3;

  gdk_draw_rgb_image (drawing_area.get_window(), 
		      drawing_area.get_style()->get_fg_gc(GTK_STATE_NORMAL),
		      e->area.x, e->area.y,
		      MIN(d.w - e->area.x ,e->area.width), 
		      MIN(d.h - e->area.y, e->area.height),
		      GDK_RGB_DITHER_MAX, pix_ptr, stride);
  return TRUE;
}

int GUIDistVis::button_release (GdkEventButton *b)
{
  switch (b->button) {
  case 1:
    play_parm->set_chunk ((int) b->x);
    printf ("Clicked %f %f\n", b->x, b->y);
    break;
  }

  return TRUE;
}

void GUIDistVis::redisp (int chunk)
{
  int prev = play_parm->get_prev (chunk);
  int sim_chunk; // chunk to which prev was similar

  if (prev == -1) // first chunk ever!
    return;

  sim_chunk = chunk == 0 ? chunk_cnt - 1 : chunk - 1;

  int age = play_parm->get_age_of_chunk (chunk);
  if (age == -1) // XXX
    return; 
  int max_age = play_parm->get_max_age ();
  draw_dat.rgbbuf [prev * chunk_cnt * 3 + sim_chunk * 3  + 0] = 
    ((max_age - age) * 255)/max_age;

  gdk_window_clear_area_e (drawing_area.get_window(), prev, sim_chunk, 1, 1);
  // or draw circles?
  // gdk_draw_arc (drawable, gc, FALSE, // not filled
  //  x, y, w, h, 0, 360 * 64)
}


DiffStore::DiffStore (guchar *p, unsigned int l)
  : pix (p), len (l), last_f (-1 * FLT_MAX), // XXX qnan would be better
  last_j (0) // just in case: should only be accessed when last_f changed
{

}

DiffStore::~DiffStore ()
{
  delete [] colors_map;
}

void DiffStore::set_len (unsigned int l)
{
  colors_map = new unsigned char[l * 3];
  for (unsigned int p = 0; p < l; p++) {
    guchar *base = colors_map + 3 * p;
    base [0] = 0; // red is reserved for indicating progress..
    base [1] = 255 * (len - p) / l;
    base [2] = 255 * p / l;
  }
}

bool DiffStore::set_diff (unsigned int which, float f, unsigned int j)
{
  if (which > len)
    return false;
  int cur_col = (f == last_f) ? last_j : j;
  guchar *cur_pix =  pix + which * 3;  
  for (int p = 0; p < 3; p++) {
    cur_pix [p] = colors_map [3 * cur_col + p];
  }
  if (last_f != f) {
    last_f = f;
    last_j = j;
  }
  return true;
}

