/* 
 * Copyright (C) 2000-1 Aaron Lav, asl2@pobox.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <vector>
#include <map>
#include <algorithm>

#include <rfftw.h>

#include <gdk/gdk.h>
#include <gdk--/window.h>
#include <gdk--/drawable.h>


#include "soundcard.h"
#include "stuff.h"
#include "vis.h"
#include "guilinvis.h"

GUILinVis::GUILinVis (PlayParm *pp) 
  : Vis (pp), mainwnd(GTK_WINDOW_TOPLEVEL)
{
  mainwnd.set_default_size(500, 500);

  drawing_area.add_events(GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);

  drawing_area.size_allocate.connect (SigC::slot(this, &GUILinVis::size_allocate));
  drawing_area.expose_event.connect (SigC::slot(this, &GUILinVis::expose));
  drawing_area.button_release_event.connect (SigC::slot
					     (this,&GUILinVis::button_release));

  mainwnd.add(drawing_area);
}

void GUILinVis::set_parms (Parm *p)
{
  if (chunk_cnt != p->nchunk) {
    chunk_cnt = p->nchunk;
    if (drawing_area.is_realized ()) {
      // there isn't a gtk-- way to get allocation?
      calc_sizes (&GTK_WIDGET(drawing_area.gtkobj())->allocation);
      drawing_area.get_window().clear_area_e (0,0,0,0);
    }
  }
  drawing_area.show ();
  mainwnd.show ();
}

void GUILinVis::redisp (int chunk)
{
  GdkRectangle r = chunk_to_rect (chunk);
  int cap_width = (chunk_height + 1) / 2;
  // Xlib Prog. Manual 5.3.1.3, diameter == line _width.
  // XXX make this more modular, so we don't distribute info 
  // about line style being CapRound all over program...

  drawing_area.get_window().clear_area_e (r.x - cap_width,
					  r.y - chunk_height/2, 
					  chunk_width + 2 * cap_width, 
					  chunk_height);
}

void GUILinVis::size_allocate (GtkAllocation *a)
{
  calc_sizes (a);
}

void GUILinVis::calc_sizes (const GtkAllocation *a)
{
  chunk_width = 10;
  chunks_per_line = a->width / chunk_width;
  if (!chunks_per_line)
    chunks_per_line = 1;
  n_lines = chunk_cnt / chunks_per_line + 1; 
  if (n_lines == 1) { // probably kinda boring...
    chunks_per_line = chunk_cnt;
    chunk_height = a->height;
  } else {
    chunk_height = a->height / n_lines;
  }
  play_parm->set_max_age (chunk_height);

  // XXX complain if chunk_height < 10?
}


GdkRectangle GUILinVis::chunk_to_rect (int i)
{
  GdkRectangle r;    
  if (!chunks_per_line) { // call before size_allocate XXX
    printf ("chunk_to_rect b/f size_allocate!\n");
    r.x = r.y = r.width = r.height = 0;
  } else {
    r.x = (i % chunks_per_line) * chunk_width;
    r.y = (i / chunks_per_line) * chunk_height + chunk_height / 2;
    r.width = chunk_width;
    r.height = chunk_height;
  }
  return r;
}

int GUILinVis::expose (GdkEventExpose *e)
{
  Gdk_Window d = drawing_area.get_window();
  Gdk_GC gc(d);


  // XXX better to calculate {i | chunk_to_rect(i) \intersect e->area != \null}

  for (unsigned int i = 0; i < chunk_cnt; i++) {
    GdkRectangle r = chunk_to_rect (i);
    GdkRectangle junk;

    // no Gdk_Rectangle::intersect in gdkmm 1.2.0??
    if (!gdk_rectangle_intersect (&r, &e->area, &junk))
      continue;
    
    int age = play_parm->get_age_of_chunk (i);
    if (age == -1)
      continue;

    int l_width = max (1, chunk_height - age - 2);
    gc.set_line_attributes (l_width, GDK_LINE_SOLID, GDK_CAP_ROUND);
    // join style is ignored, so let it default.
    d.draw_line (gc, r.x, r.y, r.x + r.width, r.y);
  }
  return TRUE;
}

int GUILinVis::button_release ( GdkEventButton *b)
{
  int line = (int) (b->y / chunk_height);

  play_parm->set_chunk ((int)
    (line * chunks_per_line + (b->x / chunk_width)));

  // XXX this will cause a skip in the middle of a chunk sometimes.  OK?
  // also, if we reset cur_playing here, we can then skip away before
  // playing the first segment specified by the user.
  return TRUE;
}



