/* 
 * Copyright (C) 2000-1 Aaron Lav, asl2@pobox.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <inttypes.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <vector>
#include <map>
#include <algorithm>

#include <rfftw.h>

#include <gtk--/main.h>

#include "soundcard.h"
#include "stuff.h"
#include "gtkplayer.h"

// duplicate notational convenience of Adjustment ctor...

void set_adj_vals (Gtk::Adjustment *adj, gfloat val, gfloat lower,
		   gfloat upper, gfloat step_incr,gfloat page_incr, 
		   gfloat page_size = 0)
{
  adj->set_lower (lower);
  adj->set_upper (upper);
  adj->set_step_increment (step_incr);
  adj->set_page_increment (page_incr);
  adj->set_page_size (page_size);
  adj->set_value (val); // set_value must be after set_{lower,upper}
}

struct {
  int digits;
  char *text;
} spin_init_info [] = {
  {0, "Number of previous segments to echo"},
  {2, "Per-segment echo attenuatation"},
  {3, "Segment duration (seconds)"}
};

GUIPlayer::GUIPlayer (int *argc, char ***argv, Samples *s, 
		      SoundPlayer *sp):
  Player (s, sp), Gtk::Main (argc, argv), mainwnd (GTK_WINDOW_TOPLEVEL),
		       jump_prob_spin(0, 2),
		       wander_check_button("Wander"),
		       recompute_button ("Recompute")
{

  jump_prob_text.set ("Jump probability");
  jump_prob_text.show ();
  set_adj_vals (jump_prob_spin.get_adjustment(),
		Player::get_jump_probability (), 0, 1, 0.01, 0.1);
  set_adj_vals (spins[echo_len_ind].spin.
		get_adjustment(),s->get_echo_len(), 0,20,1,1);
  set_adj_vals (spins[echo_amp_ind].spin.get_adjustment(),
		s->get_echo_amp(),0,1,0.01,0.1);
  set_adj_vals (spins [chunk_dur_ind].spin.get_adjustment (),
		s->get_chunk_dur(),0.005,0.5,0.010,0.1);
  vbox.show ();
  hbox.show ();
  hbox.pack_start (jump_prob_text, TRUE, FALSE, 0);
  hbox.pack_start (jump_prob_spin, TRUE, FALSE, 0);
  wander_check_button.show ();
  hbox.pack_start (wander_check_button, FALSE, FALSE, 0);
  vbox.pack_start (hbox, FALSE, FALSE, 0);
  Gtk::Main::timeout.connect (SigC::slot(this,&GUIPlayer::wander_timeout),1000);

  transition_alg_radio_buttons = new Gtk::RadioButton [transition_alg_count];
  for (int i = 0; i < transition_alg_count; i++) {
    if (i)
      transition_alg_radio_buttons[i].set_group 
	(transition_alg_radio_buttons[0].group());
    transition_alg_radio_buttons[i].add_label(transition_alg_names[i]);
    transition_alg_radio_buttons[i].show ();
    transition_alg_hbutton_box.pack_start (transition_alg_radio_buttons[i]);
    transition_alg_radio_buttons[i].
      toggled.connect(SigC::bind<int>(SigC::slot(this,&GUIPlayer::transition_alg_toggled),
				      i));
    
  }
  vbox.pack_start (transition_alg_hbutton_box, FALSE, FALSE, 0);
  mainwnd.add (vbox);
  vbox.pack_start (recompute_button, FALSE, FALSE, 0);
  recompute_button.clicked.connect(SigC::slot(this,&GUIPlayer::recompute));
  for (unsigned int j = 0; j  < max_ind; j++) {
    spins[j].box.pack_start (spins[j].text);
    spins[j].box.pack_start (spins[j].spin);
    spins[j].spin.set_digits (spin_init_info[j].digits);
    spins[j].text.set (spin_init_info[j].text);
    spins[j].text.show ();
    spins[j].spin.show ();
    spins[j].box.show ();
    vbox.pack_start (spins [j].box);
  }
  recompute_button.show ();
  transition_alg_hbutton_box.show ();
  jump_prob_spin.show ();
  vbox.show ();
  mainwnd.show ();
  Gtk::Main::input.connect(SigC::slot(this, &GUIPlayer::write_ready),
			 sound->audio_fd, GDK_INPUT_WRITE);
  bytes_left = 0;
}

float GUIPlayer::get_jump_probability ()
{
  Gtk::Adjustment *adj= jump_prob_spin.get_adjustment ();
  return 1 - adj->get_value ();
}

void GUIPlayer::transition_alg_toggled (int alg)
{
  // toggled signal is generated for all buttons w/ gtk-- 1.2.0
  // whenever a single one is toggled.
  if (transition_alg_radio_buttons[alg].get_active()) {
    transition_alg = alg;
  }
  return;
}

void GUIPlayer::recompute ()
{
  samp->set_chunk_dur (spins[chunk_dur_ind].spin.get_adjustment()
		       ->get_value ());
  samp->set_echo_len (spins[echo_len_ind].spin.get_adjustment()->get_value ());
  samp->set_echo_amp (spins[echo_amp_ind].spin.
		      get_adjustment()->get_value ());
  samp->make_chunks ();
  clear_hist ();
  Vis::Parm vp = {samp->get_chunk_cnt(), samp};
  for (set<Vis *>::const_iterator i = visualizer_set.begin();
       i != visualizer_set.end (); i++) {
    (*i)->set_parms (&vp);
  }
}

gint GUIPlayer::wander_timeout ()
{
  if (wander_check_button.get_active ()) {
    float ch = 0.01 * (1 - ((rand ()/17)  % 3));
    Gtk::Adjustment *adj = jump_prob_spin.get_adjustment ();
    float new_val = adj->get_value () + ch;
    new_val = min (new_val, adj->get_upper());
    new_val = max (new_val, adj->get_lower());
    adj->set_value (new_val);
  }
  return TRUE;
}

void GUIPlayer::add_vis (Vis *v)
{
  visualizer_set.insert (v);
}

void GUIPlayer::write_ready (gint src, GdkInputCondition condition)
{
  if (!bytes_left) {
    calc_next ();

    for (ttc_map::iterator it = time_to_chunk.begin (); 
	 it != time_to_chunk.end (); it++) {
      int chunk = it->second;
      for (set<Vis *>::iterator v = visualizer_set.begin ();
	   v != visualizer_set.end (); v++) 
	(*v)->redisp (chunk);

      // XXX should stop erase loop when we get past stuff which has been
      // played?  Maybe not, since they might be played by
      // the time the expose is processed.  Hmm.

    }

    bytes_left = samp->get_chunk_sz ()* sizeof (sample);
  }

  // XXX move back into soundcard.cc
  int cb = write (sound->audio_fd, 
		  ((unsigned char *)samp->get_chunk_dat(cur_playing)) +
		  samp->get_chunk_sz () * sizeof (sample) -bytes_left,
		  bytes_left);
  if (cb == -1) { // check for EINTR?
    perror ("Writing");
    exit (255);
  }
  bytes_left -= cb;
}

void GUIPlayer::set_chunk (unsigned int chunk)
{
  cur_playing = chunk;
}

void GUIPlayer::set_max_age (int age)
{
  max_age =  max_chunks_buf_in_OSS + age; 
// XXX set_age isn't right, since multiple sources can age...

  // needs to include both delay from sending data to soundcard to playing,
  // + time for display to age out
}

int GUIPlayer::get_max_age ()
{
  return max_age - max_chunks_buf_in_OSS;
}

int GUIPlayer::get_age_of_chunk (unsigned int chunk)
{
  int age = -1;
  unsigned int chunk_played =  sound->get_samples_played () / 
    samp->get_chunk_sz();
    
  std::map<unsigned int, unsigned int>::reverse_iterator it;

  for (it = time_to_chunk.rbegin (); it != time_to_chunk.rend (); it++) {
    if (it->second == chunk && it->first <= chunk_played) { 
      // otherwise handed to soundcard, but not yet playing, don't bother.
      age = chunk_played - it->first;
      break;
    }
  }

  if (age == -1) {
    if (play_times[chunk])
      age = chunk_played - play_times[chunk];
  }
  return age;
}

int GUIPlayer::get_prev (unsigned int chunk)
{
  return prev_chunk[chunk];
}


void GUIPlayer::run ()
{
  gtk_main ();
}

