/* 
 * Copyright (C) 2000-1 Aaron Lav, asl2@pobox.com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <audiofile.h>

#include "stuff.h"

#include "sq.h"

void die (const char *s)
{
  perror (s);
  exit (255);
}

Samples::Samples ()
{
  cs = 0;
  chunk_sz = 0;
  chunk_cnt = 0;
  window = 0;
  data = 0;
  plan = 0;
}

void Samples::set_chunk_dur (float secs)
{
  chunk_duration = secs;
}

float Samples::get_chunk_dur ()
{
  return chunk_duration;
}

void Samples::calc_window ()
{
  window = new fftw_real[chunk_sz];
  //  static const float alpha = 0.54; // 0.54 is Hamming, 0.5 is Hanning
  static const float alpha = 0.5;
  for (unsigned int i = 0; i < chunk_sz; i++)
    window [i] = alpha - (1-alpha)* cos ((M_PI * 2 * i) / (chunk_sz - 1));
  // See Rabiner and Gold, _Theory and Practice of Digital Signal Processing_,
  // or GNU Octave 2.0.16

  printf ("Window %f %f %f\n", window[0], window [chunk_sz/2], 
	  window[chunk_sz-1]);
}

void Samples::apply_window (fftw_real buf[])
{
  for (unsigned int i = 0; i < chunk_sz; i++)
    buf[i] *= window[i];
}
// we could wait to do this echoing in the frequency domain.
// That's a win when we reduce our resolution at the
// high-frequency end (compute_power_and_reduce)
//   Also, we'd avoid duplicate int-to-real conversions.

void Samples::copy_and_echo (fftw_real * buf, int c)
{
  unsigned int len = chunk_sz;
#define max(a,b) ((a) > (b) ? (a) : (b))
  int start = max (0, c - echo_len + 1);
#undef max
  float mul = 1;
  for (int t = c; t >= start; t--) {
    const sample *samp =  get_chunk_dat (t); 
    for (unsigned int i = 0; i < len; i++) {
      if (t == c)
	buf [i] = samp [i]; // OK not to multiply by mul b/c mul == 1
      else
	buf [i]+= samp [i] * mul;
    }
    mul *= echo_amp;
  }
}

void Samples::diffs_to_pixels (DiffStore *ds, unsigned int i)
{
  return cs[i].diffs_to_pixels (ds);
}

void Samples::read (const char *fn)
{
  AFfilehandle afh = afOpenFile (fn, "r", NULL);
  if (!afh)
    die ("Opening");
  d_len = afGetFrameCount (afh, AF_DEFAULT_TRACK);
  printf ("Rate %f\n", afGetRate (afh, AF_DEFAULT_TRACK));
  rate = (int) afGetRate (afh, AF_DEFAULT_TRACK);
  int frame_sz = (int) afGetFrameSize (afh, AF_DEFAULT_TRACK, 1); // expand3to4 b/c reading does too...
  int nch = afGetChannels (afh, AF_DEFAULT_TRACK);
  int samp_sz = frame_sz / nch;
  assert (frame_sz == nch * samp_sz);
  if (samp_sz != sizeof (sample)) { 
    printf ("Unsupported frame_sz %d samp_sz %d nch %d\n", frame_sz,
	    samp_sz, nch);
    exit (255);
  }
  size_t to_alloc = d_len * samp_sz * nch;
  data = (sample *) malloc (to_alloc);
  if (!data) {
    printf ("malloc of %d failed\n", to_alloc);
    exit (255);
  }

  // afReadFrames returns an int, but an OOB value like -1 will lose the 
  // comparison below anyway
  size_t ret = (size_t) afReadFrames (afh, AF_DEFAULT_TRACK, data, d_len);

  if (ret != d_len) {
    printf ("Short read %d %d\n", (int) ret, d_len);
    exit (255);
  }

  assert (sizeof (int_least32_t) > sizeof (sample));
  if (nch != 1) {
    size_t src, dest;
    for (src = 0, dest = 0; src < d_len * nch; src += nch, dest++) {
      int_least32_t mix = 0;
      for (int ch = 0; ch < nch; ch++) {
	mix += data [src + ch];
      }
      data [dest] = (sample)  mix;
    }
    data = (sample *) realloc (data, d_len * samp_sz);
    if (!data) {
      printf ("Multichannel realloc smaller failed\n");
      exit (255); // malloc arena corruption, probably...
    }
  }
  afCloseFile (afh);

}

void Samples::calc_sz ()
{
  chunk_sz = (size_t) (get_rate () * chunk_duration);
  // product of small prime factors best, see FFTW docs.  We should
  // try to round to an appropriate number.  OTOH, that's just twiddling
  // the constant in an O(chunk_cnt * chunk_sz * log (chunk_sz)) process,
  // and empirically the O(chunk_cnt^2) diff-calculation takes much
  // longer. XXX
  // XXX we could speed things up by avoiding destroying plan,
  // cs if length is same as before.  Unfortunately, that's
  // not the slow part...
  if (plan)
    rfftw_destroy_plan (plan);

  plan = rfftw_create_plan (chunk_sz, FFTW_REAL_TO_COMPLEX, FFTW_MEASURE |
			    FFTW_IN_PLACE);
  assert (plan); // if this triggers, fftw has been wackily configured,
  // one hopes by someone who can undo the wackiness
  chunk_cnt = d_len / chunk_sz;
  printf ("Chunk_cnt %d\n", get_chunk_cnt());
  printf ("Chunk sz %d\n", chunk_sz);

  if (cs)
    delete [] cs;
  cs = new Chunk [chunk_cnt];
  calc_window ();
}

const sample * Samples::get_chunk_dat (size_t i)
{
  return data + (i * chunk_sz);
}

unsigned int Samples::get_similar (float prob, unsigned int base)
{
  return cs[base].get_similar(prob);
}


void Samples::set_echo_len (unsigned int len)
{
  echo_len = len;
}
unsigned int Samples::get_echo_len ()
{
  return echo_len;
}

void Samples::set_echo_amp (float amp)
{
  echo_amp  = amp;
}

float Samples::get_echo_amp ()
{
  return echo_amp;
}

// what if there's just no power in a particular freq band? -inf is
// hard to work with....

fftw_real my_log (fftw_real l) {
  //   return l; // even quicker hack // XXX make this configureable?
  fftw_real tmp = log (l + 1); // + 1 so that small differences
  // in loudness (i.e. near 0) don't make a big neg. difference
  return tmp; 
}


void assert_not_nan (const vector<fftw_real>& v, int c) 
{
  for (unsigned int i = 0; i < v.size (); i++) {
    if  (isnan (v[i])) {
      printf ("Bogus at %d %d\n", i, c);
    }
  }
}

void compute_power (fftw_real *buf, size_t len, 
			       vector<fftw_real>& loudness, bool dbg = false)
{
  size_t loudness_len = (len) / 2;
  // len/2 + 1, but then don't store dc.

  loudness.resize (loudness_len, 0);
  for (unsigned int k = 1; k < (len +1) / 2; k++) { // limit is round up n/2 
    loudness [k - 1] = my_log (sq(buf[k]) + sq(buf [len-k]));
  }

  if (len % 2 == 0) {
    loudness [(len / 2 - 1)] = my_log (sq(buf [len/2]));
  }

  assert ((len / 2 - 1) < loudness_len);
}

// (psychoacoustically, I suspect 
// high-freq resolution is less important than low-freq:
// 256 Hz vs. 260 Hz should be as noticeable 16384 Hz vs. 16640 Hz,
// 6 octaves up.  Is there an alternate (non FFT) transformation more
// appropriate here?)

// octave is used in the sense of frequency doublings
// the first octave is mapped 1-to-1, 2nd 1-2, 3rd 1-4, 4th 1-8.
// so loudness_len = (1/4 + 1/8 + 1/16 + 1/32) * old_len ==
// (1/2 - 1/32) * old_len, or, in general
// \sigma_{i=1}^{i=octaves} (1/octaves * 1/2^{i - 1}) ==
// 1/octaves * {2 - 1/2^{octaves - 1}}, except that
// there are rounding errors!  But it's still a useful approximation.

// The following is more obviously right (to me right now...)
// round the # of freqs in octaves 1 .. {octaves - 1} down,
// and add in an extra term for the octave-th octave

void compute_power_and_reduce (fftw_real *buf, size_t len, 
			       vector<fftw_real>& loudness, 
			       unsigned int octaves, bool dbg = false)
{
  assert (len > 1);
  // since we're ignoring the DC component, it's moderately pointless 
  // to pass in an array with no elements or just a DC component.

  assert (octaves != 0);
  // Duh.

  if (octaves > 29) { // XXX should be bits in int.
    printf ("Clipping octaves to 29 from %d\n", octaves);
    octaves = 29;
    // need to clip b/c we use 1 << oct to compute 2^oct.
    // and we may increment oct below to deal w/ leftovers.
    // we could make sure that unsigned is used everywhere
    // below and get one extra bit.
  }

  size_t loudness_len = (len) / 2; // size of unreduced power array
  // len/2 + 1, but then don't store dc.

  int octave_len = loudness_len / octaves;
  size_t reduced_len = 0;
      
  int dest_starts [octaves + 1];
  dest_starts [0] = 0;
  unsigned int divisor = 1;
  for (unsigned int i = 0; i < octaves; i++, divisor = divisor << 1) {
    assert (divisor == 1U << i);
    int adding = (loudness_len/octaves + (divisor - 1)) / divisor;
    reduced_len += adding;
    if (dbg) printf ("Adding %d giving %d\n", adding, reduced_len);
    dest_starts [i + 1] = reduced_len; // if exact, don't use i+1st, but OK
  }
  int left_over = ((loudness_len % octaves) + (1 << octaves) - 1)/
    (1 << octaves);
  if (left_over)
    octaves++;
  if (dbg) printf ("Leftover %d\n", left_over);
  reduced_len += left_over;

  loudness.resize (reduced_len, 0);

  unsigned int k = 0; 
  unsigned int dest = 0; // initialize only to shut up compiler warning

  for (unsigned int cur_oct = 0; cur_oct < octaves; cur_oct++) {
    double sqrt_oct_spacing = 1 << ((cur_oct)/2);
    if ((cur_oct) % 2)
      sqrt_oct_spacing *= M_SQRT2;
    unsigned int oct_start_k = k;
    unsigned int oct_spacing = 1 << cur_oct;
    if (dbg) printf ("Oct %d of %d\n", cur_oct, octaves);

    for (k = cur_oct * octave_len; 
	 k < MIN((cur_oct + 1)*(octave_len), loudness_len);k++) {
      dest = (k - oct_start_k) / oct_spacing 
	+ dest_starts [cur_oct];
      assert (dest < reduced_len);

      if (dbg) 
	printf ("%d %d %d %d %d\n", dest, k, oct_start_k,
		dest_starts[cur_oct], cur_oct);
      fftw_real t;

      if (k == len / 2 - 1) {
	if (dbg) printf ("Last unmatched real\n");
	t = sq(buf[len/2]);
      }
      else
	t = sq(buf[k+1]) + sq(buf [len-(k+1)]);
      loudness[dest] += my_log (t) / oct_spacing; // XXX
      // also note that this really undervalues the last entry
      // per bin
      if (dbg) 
	printf ("t %f l %f sqrt_sp %f\n", t, loudness[dest],
		sqrt_oct_spacing);

      // What exactly is the right thing to do here?  log before
      // or after addition into bins?
      // We divide by sqrt_octave_spacing on the theory that
      // if loudness is equally distributed across freq bins,
      // when we go to sum squares of diffs later, we'll
      // get the same result after octave binning as before
    }
  }
  if (dest != reduced_len -1) {
    printf ("dest %d reduced_len - 1 %d", dest, reduced_len - 1);
    fflush (stdout);
    assert (0);
  }
}

void Samples::calc_loudness (int c, vector<fftw_real>& loudness)
{
  unsigned int len = get_chunk_sz();
  fftw_real buf [len];

  if (!c && false) { // XXX enable this from cmd line?
    // 0 is a known problem. 29 is largest oct we can use (see above)
    for (unsigned int my_oct = 1; my_oct < 30; my_oct++) //
      for (unsigned int i = 2; i < MIN(len, 4000); i++) {
	printf ("Testing %d %d\n", i, my_oct);
	memset (buf, 0, i * sizeof buf[0]);
	compute_power_and_reduce (buf, i, loudness, my_oct);
      }
    loudness.resize (0);
  }

  copy_and_echo (buf, c);
  apply_window (buf);
  rfftw_one (plan, buf, NULL);

  // now buf has r_0 ... r_n/2, i_{(n+1)/2 - 1}, ..., i1.
  // r_n/2 only if len % 2 = 0, otherwise no.

#if 1
  int octaves = 4;
  compute_power_and_reduce (buf, len, loudness, octaves);
#else
  compute_power (buf, len, loudness);
#endif
  assert_not_nan (loudness, c);
}


void Samples::make_chunks ()
{
  calc_sz ();

  printf ("echo len %d amp %f\n", echo_len, echo_amp);
  printf ("Calcing loudness\n");
  time_t start = time (NULL);
  for (unsigned int i = 0; i < chunk_cnt; i++) {
    calc_loudness (i, cs[i].get_loudness());
  }
  printf ("Loudness took %ld secs\n", time(NULL) - start);
  printf ("Calcing diffs\n");
  start = time (NULL);

  for (unsigned int i = 0; i < chunk_cnt; i++)
    cs[i].calc_diffs (cs, chunk_cnt, i);

  time_t diff = time (NULL) - start;
  printf ("Took %ld secs, %e secs/(cnt^2 * sz)\n", 
	  (long) diff, diff * 1.0 / (chunk_cnt * chunk_cnt *
				     chunk_sz));
}
