/* RhythmLab: a program for experimenting with rhythm
 * Copyright (C) 1999 Aaron Lav
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
#include <sys/ioctl.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <errno.h>
#include <assert.h>

#include <stdio.h>
#include "soundcard.h"

SoundPlayer::SoundPlayer ()
{
  assert (sizeof (sample) == 2);
  assert (sizeof (usample) == 2);
  assert (sizeof (sample) < sizeof (bigger_than_sample));
  debug = 0;
  req_audio_fmt = 0;
  bytes_written = 0;
  verbose = false;
}

bool SoundPlayer::create (const char *filename)
{
  bool proc_little_endian = is_little_endian ();

  if (debug) {
    printf ("Debugging\n");
    audio_fd = open (filename, O_WRONLY | O_CREAT | O_TRUNC, 0600);
    if (audio_fd == -1) {
      perror ("Opening");
      return false;
    }
    AF_to_conv_parms (req_audio_fmt ? req_audio_fmt : AFMT_S16_LE, 
		      proc_little_endian);
    return true;
  } else {
    int stereo = nchan - 1;

    if ((audio_fd = open(filename, O_WRONLY, 0)) == -1) {
      perror("Opening dsp");  
      return false;
    }  

    int audio_format;
    if (!req_audio_fmt) {
      int formats [6];
      int i = 0;
      // prefer 16-bit formats for greater fidelity. 
      // prefer signed-to-unsigned conversion over endian reversal:
      //  probably faster to implement.

      if (proc_little_endian) {
	formats [i++] = AFMT_S16_LE;
	formats [i++] = AFMT_U16_LE;
	formats [i++] = AFMT_S16_BE;
	formats [i++] = AFMT_U16_BE;
      } else {
	formats [i++] = AFMT_S16_BE;
	formats [i++] = AFMT_U16_BE;
	formats [i++] = AFMT_S16_LE;
	formats [i++] = AFMT_U16_LE;
      }

      formats [i++] = AFMT_S8;
      formats [i++] = AFMT_U8;

      
      // Ignore mu-law, a-law, IMA ADPCM, MPEG, unless systems which only 
      // have such hardware and lack software conversions exist.

      if (!try_formats (formats, i, &audio_format)) {
	printf ("No supported sound formats among:");
	for (int j = 0; j < i; j++)
	  printf (" 0x%x", formats [j]);
	printf ("\n");
	return false;
      } else if (verbose)
	printf ("Selected audio format %x\n", audio_format);
      int arg = 0x00040010; 
      // 4 fragments of size 512 bytes @ 2 bytes/sample ~= 23 msec
      // of buffering total.
      if (ioctl (audio_fd, SNDCTL_DSP_SETFRAGMENT, &arg)) {
	printf ("Error setting frag size\n");
      }
    } else {
      if (!try_formats (&req_audio_fmt, 1, &audio_format)) {
	printf ("Requested audio format 0x%x not supported\n",
		req_audio_fmt);
	return false;
      }
    }

    AF_to_conv_parms (audio_format, proc_little_endian);

    if (ioctl(audio_fd, SNDCTL_DSP_STEREO, &stereo)==-1) {
      perror("SNDCTL_DSP_STEREO");
      return false;
    }
    nchan = stereo + 1;
    if (verbose)
      printf ("number of channels is %d\n", nchan);

    if (ioctl(audio_fd, SNDCTL_DSP_SPEED, &samprate)==-1) {
      perror("SNDCTL_DSP_SPEED");  
      return false;
    }  

    int caps;

    if (ioctl (audio_fd, SNDCTL_DSP_GETCAPS, &caps)) {
      perror ("SNDCTL_DSP_GETCAPS");
    } else {
      if (caps & DSP_CAP_BATCH)
	printf ("Display may be rather desynchronized: sound driver DSP_CAP_BATCH set\n");
      else if (!(caps & DSP_CAP_REALTIME))
	printf ("Display may be a bit desynchronized: sound driver DSP_CAP_REALTIME not set\n");
      // even so, it's more likely to be OK than just using
      // bytes_written.
      // We could actually, in the last case, calculate the
      // maximum inaccuracy by dividing fragment size by sampling rate.
      // any point?
    }
    if (verbose) {
      printf ("Sampling rate of %d selected\n", samprate);
    }
    
    return true;
  }
}

bool SoundPlayer::try_formats (int *formats, int max, int *audio_format)
{
  // we use a loop and SNDCTL_DSP_SETFMT instead of SNDCTL_DSP_GETFMTS
  // because SNDCTL_DSP_GETFMTS only returns directly-hardware-supported 
  // formats.  Since we do all manipulation in our internal format,
  // it seems to make just as much sense to let the kernel convert
  // as it does to make us (especially since the kernel can be better
  // optimized to, for example, use special byte-swap or multimedia
  // instructions).  A shared library would be optimal (since the code
  // can be shared among multiple applications and heavily optimized
  // like the kernel; and it can run in user mode instead of kernel mode)
  // but that doesn't seem to be on offer.

  // If we ever decide to fuse the conversion step into the sample generation,
  // the above analysis stops holding, and we should first try 
  // SNDCTL_DSP_FORMATS, and fall back to this code in the event
  // that we don't understand any of the natively supported formats, 
  // but there exist software conversions.  (For example, the h/w only 
  // supports 32 samples, and the OSS API has been extended to report 
  // 32-bit formats so it makes sense to report only those.)
 
  for (int i = 0; i < max; i++) {
    int try_fmt = formats [i];
    if (verbose)
      printf ("Trying %d\n", try_fmt);
    if (ioctl(audio_fd, SNDCTL_DSP_SETFMT, &try_fmt)==-1) {
      perror("SNDCTL_DSP_SETFMT");  
      exit(2);  
    }
    if (try_fmt == formats [i]) {
      *audio_format = try_fmt;
      return true;
    } else if (verbose)
      printf ("Audio format %d not supported, driver suggests %d\n", 
	      formats[i], try_fmt);
  }  
  return false;
}

void SoundPlayer::AF_to_conv_parms (int audio_format, bool proc_little_endian)
{
  if (audio_format == AFMT_S8 || audio_format == AFMT_U8)
    conv = trunc_8;
  else {
    bool card_little_endian = audio_format == AFMT_S16_LE || 
      audio_format == AFMT_U16_LE;
    conv = (card_little_endian ^ proc_little_endian) ? rev_endian : none;
  }

  switch (conv) {
  case rev_endian:
    sample_sz = sizeof (sample);
    break;
  case trunc_8:
    sample_sz = sizeof (char);
    break;
  case none:
    sample_sz = sizeof (sample);
    break;
  }
  if (audio_format == AFMT_U8 || audio_format == AFMT_U16_LE ||
      audio_format == AFMT_U16_BE) {
    signed_to_player_offset = 1 << (sample_sz * 8 - 1);
  } else {
    signed_to_player_offset = 0;
  }
  if (verbose)
    printf ("Conversion %d signed_to_sample_offset %d sample_sz %d\n", 
	    (int) conv, signed_to_player_offset, sample_sz);
}

bool SoundPlayer::is_little_endian ()
{
  short s = 1;
  bool ret = 1 == * (char *) &s;
  if (verbose)
    printf ("Processor is %s-endian\n", ret ? "little" : "big");
  return ret;
}

SoundPlayer::~SoundPlayer ()
{
  if (0 != close (audio_fd))
    printf ("Error closing audio fd %d\n", errno);
}

int SoundPlayer::get_samples_played  ()
{
  size_t b;
  if (debug) {
    b = bytes_written; 
  } else {
    count_info inf;
    static bool error_reported = false;
    if (error_reported || ioctl (audio_fd, SNDCTL_DSP_GETOPTR, &inf)) {
      perror ("ioctl SNDCTL_GETOPTR");
      error_reported = true;
      b = bytes_written;
    } else
      b = inf.bytes;
  }
  return b /  (sample_sz * nchan); 
}

int SoundPlayer::get_samples_writeable  ()
{
  size_t b = 4096  * (sample_sz * nchan); // XXX think about this...;
  if (!debug) {
    audio_buf_info inf;
    static bool error_reported = false;
    if (error_reported || ioctl (audio_fd, SNDCTL_DSP_GETOSPACE, &inf)) {
      perror ("ioctl SNDCTL_GETOPTR");
      error_reported = true;
    } else
      b = inf.bytes;
  }
  return b /  (sample_sz * nchan); 
}


// It would be nice if there were a better way to indicate what I
// wanted: separate versions of xform_buf compiled for all 6 permutations
// of {none, rev_endian, trunc_8} x {none, signed->unsigned}, with
// none x none being no code.  It's even harder for the compiler to see
// now that signed_to_card_offset is general.  Do with templates?

// Fortunately, none of our conversions expand the data, so xform_buf
// doesn't need to do allocation.

inline void SoundPlayer::xform_buf
    (const sample *s, size_t sample_cnt, sample *alt_buf)
{
  for (unsigned int i = 0; i <  sample_cnt * nchan; i++) {
    switch (conv) {
    case rev_endian: 
      {
	usample samp = (usample) (s[i]);
	alt_buf[i] =(sample) (((samp & 0xFF) <<  8) | 
	  ((samp & 0xFF00) >> 8) + signed_to_player_offset);
	break;
      }
    case trunc_8:
      ((char *) alt_buf) [i] = (char) ((s[i] >> 8) + (char) signed_to_player_offset);
      break;
    case none:
      alt_buf[i] = s[i] + signed_to_player_offset;
      break;
    }
  }
}

/* If a signal (e.g. SIGSTOP from c-z at keyboard) is delivered,
   our write will be interrupted and return w/ partial success.  Retry.*/

int safe_write (int fd, const void *b, size_t sz)
{
  int r = 0; // if r not initialized here, gratuitous use b/f init warning? (egcs 2.91.66?
  const char *buf = (const char *) b;

  while (sz) {
    r = write (fd, buf, sz);
    if (r == -1)
      return r;
    sz -= r;
    buf += r;
  }
  return 0;
}

void SoundPlayer::write (const sample *s, size_t sample_cnt)
{
  audio_buf_info inf;
  if (ioctl (audio_fd, SNDCTL_DSP_GETOSPACE, &inf) == -1) {
    perror ("Getting buf info");
  } else {
    printf ("Writing: frags writeable %d total %d fragsize %d bytes %d\n",
	    inf.fragments, inf.fragstotal, inf.fragsize, inf.bytes);
  }
  
  if (conv != none || signed_to_player_offset) {
    sample alt_buf[sample_cnt];
    xform_buf (s, sample_cnt,alt_buf);
    finish_write (alt_buf, sample_cnt);
  } else 
    finish_write (s, sample_cnt);
}

void SoundPlayer::finish_write (const void *s, size_t sample_cnt)
{
  size_t bytesz = sample_cnt * sample_sz * nchan;
  bytes_written += bytesz;

  if (-1 == safe_write (audio_fd, s, bytesz)) {
    perror ("Writing sound");
    exit (255);
  }
}




