#ifndef SOUNDCARD_H_INC
#define SOUNDCARD_H_INC

#include <inttypes.h>


typedef int16_t sample; // XXX use autoconf!
// asserts in SoundPlayer::SoundPlayer try to guarantee that
// sizeof (sample) == sizeof (usample) < sizeof (bigger_than_sample)
typedef int   bigger_than_sample;
typedef unsigned short usample;


// at some point, SoundPlayer should be abstracted into an abstract base
// class, with implementations for OSS and some network sound 
// protocol.  If we want to get format conversions inlined into the
// mixing loop, then supported conversions have to be in the base class.
// (Everything else is per-beat or less frequent, so a virtual function
// call isn't disastrous overhead).


class SoundPlayer {
 public:
  SoundPlayer ();
  ~SoundPlayer ();

  // must call request_* before create 
  void request_samprate (int s) {samprate = s;};
  void request_nchan (int n) {nchan = n;};
  void request_debug (bool d) {debug = d;};
  void request_format (int fmt) {req_audio_fmt = fmt;};
  void request_verbose (bool v) {verbose = v;};
  bool create (const char *filename);

  // Because of lack of locking here, these can't be changed once we
  // start, and SoundPlayer object must be created before any threads which
  // need to use it are.  If you change this, beware of deadlocks
  // (Seq<...>::samples_to_msec is called from sched.cc with its lock held,
  // but the usual locking hierarchy is 1st main lock, then sched.
  int get_samprate () const {return samprate;};
  int get_nchan () const {return nchan;};
  int get_samples_played (); // caller must ensure lock wrt write ()
  void write (const sample *s, size_t sample_cnt);
  int get_samples_writeable (); // how many can we write?
  int audio_fd;
 protected:
  int bytes_written;
 private:
  void finish_write (const void *s, size_t sample_cnt);
  bool is_little_endian ();
  bool try_formats (int *formats, int max, int *audio_format);
  typedef enum {none, rev_endian, trunc_8} conv_t;
  void xform_buf (const sample *s, size_t sample_cnt, sample *alt_buf);
  void AF_to_conv_parms (int audio_format, bool proc_little_endian); 
  int samprate;
  int nchan;
  bool supports_getoptr;

  conv_t conv;
  int signed_to_player_offset; // 0 if eventual player format is signed, 
  // 2^{n-1} for n-bit unsigned
  size_t sample_sz;
  // next 3 only used during setup.  We really ought to,
  // instead, have a parameters object which is created
  // by the client, then passed to create() and discarded.
  bool debug;
  bool verbose;
  int req_audio_fmt; // only used during setup
};

#endif
